"""
State
======
Each State will hold a single ZoneMinder State.
It is basically a bunch of getters for each access to event data.
If you don't see a specific getter, just use the generic get function to get
the full object
"""


from pyzm.helpers.Base import Base

class State(Base):
    def __init__(self, api=None, state=None, logger=None):
        super().__init__(logger)
        self.state = state
        self.api = api
    
    def get(self):
        """Returns raw state object
        
        Returns:
            :class:`pyzm.helpers.State`: raw state object
        """
        return self.state['State']
    

    def active(self):
        """whether this state is active or not
        
        Returns:
            bool: True if active
        """
        return self.state['State']['IsActive'] == '1'

    def definition(self):
        """Returns the description text of this state
        
        Returns:
            string: description
        """
        return self.state['State']['Definition'] or None
    
    def name(self):
        """Name of this state
        
        Returns:
            string: name of this state
        """
        return self.state['State']['Name']
    
    def id(self):
        """Id of this state
        
        Returns:
            int: id of this state
        """
        return int(self.state['State']['Id'])

   



"""
States
=======
Holds a list of States for a ZM configuration
Given states are fairly static, maintains a cache of states
which can be overriden 
"""


from pyzm.helpers.State import State
from pyzm.helpers.Base import Base
import requests

class States(Base):
    def __init__(self,logger=None, api=None):
        super().__init__(logger)
        self.api = api
        self._load()

    def _load(self,options={}):
        self.logger.Debug(2,'Retrieving states via API')
        url = self.api.api_url +'/states.json'
        r = self.api._make_request(url=url)
        states = r.get('states')
        self.states = []
        for state in states:
           self.states.append(State(state=state,api=self.api, logger=self.logger))


    def list(self):
        """Returns list of state objects
        
        Returns:
            list of `pyzm.helpers.State`: list of state objects
        """
        return self.states

    
    def find(self, id=None, name=None):
        """Return a state object that matches either and id or a
        
        Args:
            id (int, optional): Id of state. Defaults to None.
            name (string, optional): name of state. Defaults to None.
        
        Returns:
            :class:`pyzm.helpers.State`: State object that matches
        """
        if not id and not name:
            return None
        match = None
        if id:
            key = 'Id'
        else:
            key = 'Name'
    
        for state in self.states:
            if id and state.id() == id:
                match = state
                break
            if name and state.name().lower() == name.lower():
                match = state
                break
        return match
        


    




"""
ZMEventNotification
=====================
Implements a python implementation of the ZM ES server. 

"""

import websocket
import json
import time
from pyzm.helpers.Base import Base
import threading
import ssl

class ZMEventNotification(Base):
    def __init__(self, options):
        """Instantiates a thread that connects to the ZM Notification Server

        Args:
            options (dict): As below::
        
                {
                    'url': string # websocket url
                    'user': string # zm user name
                    'password': string # zm password
                    'allow_untrusted': boolean # set to true for self-signed certs
                    'on_es_message': callback function when a message is received
                    'on_es_close': callback function when the connection is closed
                    'on_es_error': callback function when an error occurs
                }
        
        Raises:
            ValueError: if no server is provided
        """

        super().__init__(options.get('logger'))
        if not options.get('url'):
            raise ValueError ('ZMESClient: No server specified')
        
        self.url = options.get('url')
        self.user = options.get('user')
        self.password = options.get('password')
        self.allow_untrusted = options.get('allow_untrusted')

        self.on_es_message = options.get('on_es_message')
        self.on_es_close = options.get('on_es_close')
        self.on_es_error = options.get('on_es_error')

        self.ws = None
        self.queue = []

        self.connected = False
        self.connect()
    
    def connect(self):
        """Connect to the ES
        """
        self.logger.Info('ZMESClient: Connecting to ES')
        if not self.connected:
            self._disconnect = False
            self.worker_thread = threading.Thread(target=self._worker)
            self.worker_thread.start()
            self.logger.Info ('ZMESClient: Event Server init started')
    
    def disconnect(self):
        """Disconnect from the ES
        """
        self.logger.Info('ZMESClient: Disconnecting from ES')
        if self.ws != None:
            self._disconnect = True
            self.ws.keep_running = False
            if self.worker_thread != None:
                self.worker_thread.join(15)
        # We have waited for the thread to finish
        # We are now disconnected
        self.connected = False
        self._disconnect = False
            
    def send(self, msg):
        """Send message to ES
        
        Args:
            msg (dict): message to send. The message should follow a control structure as specified in The `ES developer guide`_

        .. _ES developer guide:
                https://zmeventnotification.readthedocs.io/en/latest/guides/developers.html

        """
        if self.connected:
            self.ws.send(json.dumps(msg))
        else:
            self.logger.Debug (1,'ZMESClient: not yet connected, message queued[{}]: {}'.format(len(self.queue), msg))
            self.queue.append(msg)


    def _monkey_callback(self,callback, *args):
        """
        Monkey patch for WebSocketApp._callback() because it swallows
        exceptions.
        """
        if callback is not None:
            callback(self.ws, *args)
    
    def _worker(self):
        self.logger.Info('ZMESClient: Inside Event Server thread, attempting to connect')
        sslopt = {}
        if self.allow_untrusted:
            sslopt['cert_reqs'] = ssl.CERT_NONE
            self.logger.Warning('ZMESClient: Turning off certificate trust')

        while not self._disconnect:
            self.ws = websocket.WebSocketApp(self.url, 
                                        on_message = lambda ws,msg:  self._on_message(ws, msg), 
                                        on_error = lambda ws,msg:  self._on_error(ws,msg), 
                                        on_close = lambda ws:  self._on_close(ws),
                                        on_open = lambda ws: self._on_open(ws)
                                        )
            self.ws._callback  = self._monkey_callback
            self.logger.Info ('ZMESClient: connected: ready to send/receive websocket messages')
            try:
                val = self.ws.run_forever(sslopt=sslopt)
                if not val: break # keyboard
            except Exception as e:
                self.logger.Error ('ZMESClient: Event Server Exception:' + str(e))
                
                #traceback.print_exc(file=sys.stdout)

            # The connection is aborted (intential or not)
            self.connected = False
            self.ws.close()
            self.ws = None
                        
            if not self._disconnect:
                self.logger.Error ('ZMESClient: run_forever() unexpectedly terminated' )
                self.logger.Info('ZMESClient: Will reconnect after 10 secs...')
                time.sleep(10)

        # Not connected anymore
        self.logger.Info('ZMESClient: Exiting Event Server thread, correctly disconnected')
        
    def _on_open(self, ws):   
        self.logger.Info('ZMESClient: Sending auth info to ES')
        auth={"event":"auth","data":{"user":self.user,"password":self.password}}
        dauth={"event":"auth","data":{"user":self.user,"password":'****'}}

        self.logger.Debug(1, 'ZMESClient: Auth info to be sent: {}'.format(dauth))
        ws.send(json.dumps(auth))

        

    def _on_message(self, ws, message):
        self.logger.Info('ZMESClient: Got message from ES: {}'.format(message))
        message = json.loads(message)
        if message.get('event') == 'auth' and message.get('status') == 'Success':
            self.logger.Info ('ZMESClient: Auth accepted, connected state')
            self.connected = True
            while self.queue:
                msg = self.queue.pop(0)
                self.logger.Debug (1, 'Sending pending message:{}'.format(msg))
                self.send(msg)

        if self.on_es_message: self.on_es_message(message)

    def _on_error(self, ws, error):
        self.logger.Error('ZMESClient: Got error: {}'.format(error))
        if self.on_es_error: 
            self.logger.Info('invoking app error function and re-raising error')
            self.on_es_error(error)
            self.ws.close()
        raise error
        
      

    def _on_close(self, ws):
       self.logger.Info ('ZMESClient: Connection closed')
       if self.on_es_close: self.on_es_close()

    



    



"""
Monitor
=======
Each Monitor will hold a single ZoneMinder Monitor.
It is basically a bunch of getters for each access to event data.
If you don't see a specific getter, just use the generic get function to get
the full object
"""


from pyzm.helpers.Base import Base

class Monitor(Base):
    def __init__(self, api=None, monitor=None, logger=None):
        super().__init__(logger)
        self.monitor = monitor
        self.api = api
    
    def get(self):
        """Returns monitor object
        
        Returns:
            :class:`pyzm.helpers.Monitor`: Monitor object
        """
        return self.monitor['Monitor']

    def enabled(self):
        """True if monitor is enabled
        
        Returns:
            bool: Enabled or not
        """
        return self.monitor['Monitor']['Enabled'] == '1'

    def function(self):
        """returns monitor function
        
        Returns:
            string: monitor function
        """
        return self.monitor['Monitor']['Function']
    
    def name(self):
        """Returns monitor name
        
        Returns:
            string: monitor name
        """
        return self.monitor['Monitor']['Name']
    
    def id(self):
        """Returns monitor Id
        
        Returns:
            int: Monitor Id
        """
        return int(self.monitor['Monitor']['Id'])

    def type(self):
        """Returns monitor type
        
        Returns:
            string: Monitor type
        """
        return self.monitor['Monitor']['Type']
    
    def dimensions(self):
        """Returns width and height of monitor
        
        Returns:
            dict: as below::

            {
                'width': int,
                'height': int
            }
        """
        return { 'width':int(self.monitor['Monitor']['Width']), 
                 'height':int(self.monitor['Monitor']['Height'])
        }
    
    def events(self, options={}):
        """Returns events associated to the monitor, subject to filters in options
        
        Args:
            options (dict, optional): Same as options for :class:`pyzm.helpers.Event`. Defaults to {}.
        
        Returns:
           :class:`pyzm.helpers.Events`
        """
        options['mid'] = self.id()
        return self.api.events(options=options)

    def eventcount(self, options={}):
        """Returns count of events for monitor
        
        Args:
            options (dict, optional): Same as options for :class:`pyzm.helpers.Event`. Defaults to {}.
        
        Returns:
            int: count
        """
        # regular events API is more powerful than
        # console events as it allows us flexible timings
        # making limit=1 keeps the processing limited
        options['mid'] = self.id()
        options['max_events'] = 1
        return self.api.events(options=options).count()
        #print (s)
    
    def delete(self):
        """Deletes monitor
        
        Returns:
            json: API response
        """
        url = self.api.api_url+'/monitors/{}.json'.format(self.id())
        return self.api._make_request(url=url, type='delete')


    def set_parameter(self, options={}):
        """Changes monitor parameters
        
        Args:
            options (dict, optional): As below. Defaults to {}::

                {
                    'function': string # function of monitor
                    'name': string # name of monitor
                    'enabled': boolean
                    'raw': {
                        # Any other monitor value that is not exposed above. Example:
                        'Monitor[Colours]': '4',
                        'Monitor[Method]': 'simple'
                    }

                }
    
        
        Returns:
            json: API Response
        """
        url = self.api.api_url+'/monitors/{}.json'.format(self.id())
        payload = {}
        if options.get('function'):
            payload['Monitor[Function]'] = options.get('function')
        if options.get('name'):
            payload['Monitor[Name]'] = options.get('name')
        if options.get('enabled') != None:
            enabled = '1' if options.get('enabled') else '0'
            payload['Monitor[Enabled]'] = enabled

        if options.get('raw'):
            for k in options.get('raw'):
                payload[k] = options.get('raw')[k]
               
        if payload:
            return self.api._make_request(url=url, payload=payload, type='post')

    def arm(self):
        """Arms monitor (forces alarm)
        
        Returns:
            json: API response
        """
        return self._set_alarm(type='on')

    def disarm(self):
        """Disarm monito (removes alarm)
        
        Returns:
            json: API response
        """
        return self._set_alarm(type='off')

    def _set_alarm(self,type='on'):
        url = self.api.api_url+'/monitors/alarm/id:{}/command:{}.json'.format(self.id(), type)
        return self.api._make_request(url=url)



    def status(self):
        """Returns status of monitor, as reported by zmdc
            TBD: crappy return, need to normalize
        
        Returns:
            json: API response
        """
        url = self.api.api_url+'/monitors/daemonStatus/id:{}/daemon:zmc.json'.format(self.id())
        return self.api._make_request(url=url)



"""
Base
======
All classes derive from this Base class.
It implements some common functions that apply across all.
For now, this basically holds a pointer to the logging function to invoke
to log messages including a simple console based print function if none is provided

"""

from datetime import datetime

class Base:
    def __init__(self, logger):
        #print ('core:logger is {}'.format(logger))
        if not logger:
            self.logger = ConsoleLog()
            self.logger.Info ('Using simple log output (default)')
            
        else:
            self.logger = logger
    
    def get_logger(self):
        return self.logger

class ConsoleLog:
    ' console based logging function that is used if no logging handler is passed'
    def __init__(self):
        self.dtformat = "%b %d %Y %H:%M:%S.%f"
        self.level = 5

    def set_level(self,level):
        self.level = level

    def get_level(self):
        return self.level

    def Debug (self,level, message, caller=None):
        if level <= self.level:
            dt = datetime.now().strftime(self.dtformat)
            print ('{} [DBG {}] {}'.format(dt, level, message)) 

    def Info (self,message, caller=None):
        dt = datetime.now().strftime(self.dtformat)
        print ('{} [INF] {}'.format( dt, message))

    def Warning (self,message, caller=None):
        dt = datetime.now().strftime(self.dtformat)
        print ('{}  [WAR] {}'.format( dt, message))

    def Error (self,message, caller=None):
        dt = datetime.now().strftime(self.dtformat)
        print ('{} [ERR] {}'.format(dt, message))

    def Fatal (self,message, caller=None):
        dt = datetime.now().strftime(self.dtformat)
        print ('{} [FAT] {}'.format(dt, message))
        exit(-1)

    def Panic (self,message, caller=None):
        dt = datetime.now().strftime(self.dtformat)
        print ('{} [PNC] {}'.format(dt, message))
        exit(-2)



import numpy as np

import sys
import cv2
import time
import datetime
import re
from pyzm.helpers.Base import Base
import portalocker
import os
from pyzm.helpers.utils import Timer
# Class to handle Yolo based detection


class Yolo(Base):

    # The actual CNN object detection code
    # opencv DNN code credit: https://github.com/arunponnusamy/cvlib

    def __init__(self, options={}, logger=None):
        super().__init__(logger)
        self.net = None
        self.classes = None
        self.options = options
        self.is_locked = False

        #self.logger.Debug (4, 'Yolo init params: {}'.format(options))

        self.processor=self.options.get('object_processor') or 'cpu'
        self.lock_maximum=int(options.get(self.processor+'_max_processes') or 1)
        self.lock_timeout = int(options.get(self.processor+'_max_lock_wait') or 120)
        
        #self.lock_name='pyzm_'+self.processor+'_lock'
        self.lock_name='pyzm_uid{}_{}_lock'.format(os.getuid(),self.processor)

        self.disable_locks = options.get('disable_locks', 'no')
        if self.disable_locks == 'no':
            self.logger.Debug (2,f'portalock: max:{self.lock_maximum}, name:{self.lock_name}, timeout:{self.lock_timeout}')
            self.lock = portalocker.BoundedSemaphore(maximum=self.lock_maximum, name=self.lock_name,timeout=self.lock_timeout)
        self.model_height = self.options.get('model_height', 416)
        self.model_width = self.options.get('model_width', 416)

    def acquire_lock(self):
        if self.disable_locks=='yes':
            return
        if self.is_locked:
            self.logger.Debug(2, '{} portalock already acquired'.format(self.lock_name))
            return
        try:
            self.logger.Debug (2,f'Waiting for {self.lock_name} portalock...')
            self.lock.acquire()
            self.logger.Debug (2,f'Got {self.lock_name} portalock')
            self.is_locked = True
           
        except portalocker.AlreadyLocked:
            self.logger.Error ('Timeout waiting for {} portalock for {} seconds'.format(self.lock_name, self.lock_timeout))
            raise ValueError ('Timeout waiting for {} portalock for {} seconds'.format(self.lock_name, self.lock_timeout))


    def release_lock(self):
        if self.disable_locks=='yes':
            return
        if not self.is_locked:
            self.logger.Debug (2, '{} portalock already released'.format(self.lock_name))
            return
        self.lock.release()
        self.is_locked = False
        self.logger.Debug (2,'Released {} portalock'.format(self.lock_name))


        
    def populate_class_labels(self):
        class_file_abs_path = self.options.get('object_labels')
        f = open(class_file_abs_path, 'r')
        self.classes = [line.strip() for line in f.readlines()]
        f.close()

    def get_classes(self):
        return self.classes

    def load_model(self):
        self.logger.Debug (1, '|--------- Loading Yolo model from disk ---------|')
        t = Timer()
        self.net = cv2.dnn.readNet(self.options.get('object_weights'),
                                self.options.get('object_config'))
        #self.net = cv2.dnn.readNetFromDarknet(config_file_abs_path, weights_file_abs_path)
        diff_time = t.stop_and_get_ms()

        self.logger.Debug(
            1,'perf: processor:{} Yolo initialization (loading {} model from disk) took: {}'
            .format(self.processor, self.options.get('object_weights'), diff_time))
        if self.processor == 'gpu':
            (maj, minor, patch) = cv2.__version__.split('.')
            min_ver = int(maj + minor)
            if min_ver < 42:
                self.logger.Error('Not setting CUDA backend for OpenCV DNN')
                self.logger.Error(
                    'You are using OpenCV version {} which does not support CUDA for DNNs. A minimum of 4.2 is required. See https://www.pyimagesearch.com/2020/02/03/how-to-use-opencvs-dnn-module-with-nvidia-gpus-cuda-and-cudnn/ on how to compile and install openCV 4.2'
                    .format(cv2.__version__))
                self.processor = 'cpu'
        else:
            self.logger.Debug (1, 'Using CPU for detection')

        if self.processor == 'gpu':
            self.logger.Debug( 2,'Setting CUDA backend for OpenCV')
            self.logger.Debug( 3,'If you did not set your CUDA_ARCH_BIN correctly during OpenCV compilation, you will get errors during detection related to invalid device/make_policy')
            self.net.setPreferableBackend(cv2.dnn.DNN_BACKEND_CUDA)
            self.net.setPreferableTarget(cv2.dnn.DNN_TARGET_CUDA)
        
        self.populate_class_labels()

    def get_output_layers(self):
        layer_names = self.net.getLayerNames()
        output_layers = [
            layer_names[i[0] - 1] for i in self.net.getUnconnectedOutLayers()
        ]
        return output_layers

    def detect(self, image=None):
        Height, Width = image.shape[:2]
        downscaled =  False
        upsize_xfactor = None
        upsize_yfactor = None
        max_size = self.options.get('max_size', Width)
        old_image = None

        if Width > max_size:
            downscaled = True
            self.logger.Debug (2, 'Scaling image down to max size: {}'.format(max_size))
            old_image = image.copy()
            image = imutils.resize(image,width=max_size)
            newHeight, newWidth = image.shape[:2]
            upsize_xfactor = Width/newWidth
            upsize_yfactor = Height/newHeight        


        if self.options.get('auto_lock',True):
            self.acquire_lock()

        try:
            if not self.net:
                self.load_model()

            self.logger.Debug(
                1,'|---------- YOLO (input image: {}w*{}h, model resize dimensions: {}w*{}h) ----------|'
                .format(Width, Height, self.model_width, self.model_height))

            
            scale = 0.00392  # 1/255, really. Normalize inputs.
                
            t = Timer()
            ln = self.net.getLayerNames()
            ln = [ln[i[0] - 1] for i in self.net.getUnconnectedOutLayers()]
            blob = cv2.dnn.blobFromImage(image,
                                        scale, (self.model_width, self.model_height), (0, 0, 0),
                                        True,
                                        crop=False)
            
            self.net.setInput(blob)
            outs = self.net.forward(ln)

            if self.options.get('auto_lock',True):
                self.release_lock()
        except:
            if self.options.get('auto_lock',True):
                self.release_lock()
            raise

        diff_time = t.stop_and_get_ms()
        self.logger.Debug(
            1,'perf: processor:{} Yolo detection took: {}'.format(self.processor, diff_time))

    
        class_ids = []
        confidences = []
        boxes = []

        nms_threshold = 0.4
        conf_threshold = 0.2

        # first nms filter out with a yolo confidence of 0.2 (or less)
        if float(self.options.get('object_min_confidence')) < conf_threshold:
            conf_threshold = float(self.options.get('object_min_confidence'))

        for out in outs:
            for detection in out:
                scores = detection[5:]
                class_id = np.argmax(scores)
                confidence = scores[class_id]
                center_x = int(detection[0] * Width)
                center_y = int(detection[1] * Height)
                w = int(detection[2] * Width)
                h = int(detection[3] * Height)
                x = center_x - w / 2
                y = center_y - h / 2
                class_ids.append(class_id)
                confidences.append(float(confidence))
                boxes.append([x, y, w, h])

        t = Timer()
        indices = cv2.dnn.NMSBoxes(boxes, confidences, conf_threshold,
                                   nms_threshold)
        diff_time = t.stop_and_get_ms()

        self.logger.Debug(
            2,'perf: processor:{} Yolo NMS filtering took: {}'.format(self.processor, diff_time))

        bbox = []
        label = []
        conf = []

        prefix = '(yolo) ' if self.options.get('show_models')=='yes' else ''

        # now filter out with configured yolo confidence, so we can see rejections in log
        for i in indices:
            i = i[0]
            box = boxes[i]
            x = box[0]
            y = box[1]
            w = box[2]
            h = box[3]

            
            bbox.append([
                int(round(x)),
                int(round(y)),
                int(round(x + w)),
                int(round(y + h))
            ])
            label.append(prefix+str(self.classes[class_ids[i]]))
            conf.append(confidences[i])
           
        if downscaled:
            self.logger.Debug (2,'Scaling image back up to {}'.format(Width))
            image = old_image
            for box in bbox:
                box[0] = round (box[0] * upsize_xfactor)
                box[1] = round (box[1] * upsize_yfactor)
                box[2] = round (box[2] * upsize_xfactor)
                box[3] = round (box[3] * upsize_yfactor)
        
        return bbox, label, conf
